"""
.. module: historical.vpc.poller
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
"""
import os
import logging

from botocore.exceptions import ClientError

from raven_python_lambda import RavenLambdaWrapper
from cloudaux.aws.ec2 import describe_vpcs

from historical.constants import POLL_REGIONS, HISTORICAL_ROLE
from historical.vpc.models import vpc_polling_schema
from historical.common.accounts import get_historical_accounts
from historical.common.kinesis import produce_events

logging.basicConfig()
log = logging.getLogger("historical")
log.setLevel(logging.INFO)


@RavenLambdaWrapper()
def handler(event, context):
    """
    Historical VPC event poller.

    This poller is run at a set interval in order to ensure that changes do not go undetected by historical.

    Historical pollers generate `polling events` which simulate changes. These polling events contain configuration
    data such as the account/region defining where the collector should attempt to gather data from.
    """
    log.debug('Running poller. Configuration: {}'.format(event))

    for account in get_historical_accounts():
        for region in POLL_REGIONS:
            try:
                vpcs = describe_vpcs(
                    account_number=account['id'],
                    assume_role=HISTORICAL_ROLE,
                    region=region
                )

                events = [vpc_polling_schema.serialize(account['id'], v) for v in vpcs]
                produce_events(events, os.environ.get('HISTORICAL_STREAM', 'HistoricalVPCPollerStream'))
                log.debug('Finished generating polling events. Account: {}/{} '
                          'Events Created: {}'.format(account['id'], region, len(events)))
            except ClientError as e:
                log.warning('Unable to generate events for account/region. Account Id/Region: {account_id}/{region}'
                            ' Reason: {reason}'.format(account_id=account['id'], region=region, reason=e))



"""
.. module: historical.s3.collector
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
"""
import logging
from itertools import groupby

from botocore.exceptions import ClientError
from pynamodb.exceptions import PynamoDBConnectionError
from raven_python_lambda import RavenLambdaWrapper
from cloudaux.orchestration.aws.s3 import get_bucket

from historical.constants import HISTORICAL_ROLE, CURRENT_REGION
from historical.common import cloudwatch
from historical.common.kinesis import deserialize_records
from historical.s3.models import CurrentS3Model

logging.basicConfig()
log = logging.getLogger('historical')
log.setLevel(logging.INFO)


UPDATE_EVENTS = [
    "DescribeBucket",   # Polling event
    "DeleteBucketCors",
    "DeleteBucketLifecycle",
    "DeleteBucketPolicy",
    "DeleteBucketReplication",
    "DeleteBucketTagging",
    "DeleteBucketWebsite",
    "CreateBucket",
    "PutBucketAcl",
    "PutBucketCors",
    "PutBucketLifecycle",
    "PutBucketPolicy",
    "PutBucketLogging",
    "PutBucketNotification",
    "PutBucketReplication",
    "PutBucketTagging",
    "PutBucketRequestPayment",
    "PutBucketVersioning",
    "PutBucketWebsite"
]


DELETE_EVENTS = [
    "DeleteBucket",
]


def group_records_by_type(records):
    """Break records into two lists; create/update events and delete events."""
    update_records, delete_records = [], []
    for r in records:

        # Do not capture error events:
        if not r["detail"].get("errorCode"):
            if r['detail']['eventName'] in UPDATE_EVENTS:
                update_records.append(r)
            else:
                delete_records.append(r)

    return update_records, delete_records


def create_delete_model(record):
    """Create an S3 model from a record."""
    arn = "arn:aws:s3:::{}".format(cloudwatch.filter_request_parameters('bucketName', record))
    log.debug('Deleting Dynamodb Records. Hash Key: {arn}'.format(arn=arn))

    data = {
        'arn': arn,
        'principalId': cloudwatch.get_principal(record),
        'userIdentity': cloudwatch.get_user_identity(record),
        'accountId': record['account'],
        'eventTime': record['detail']['eventTime'],
        'BucketName': cloudwatch.filter_request_parameters('bucketName', record),
        'Region': cloudwatch.get_region(record),
        'Tags': {},
        'configuration': {},
        'eventSource': record["detail"]["eventSource"]
    }

    return CurrentS3Model(**data)


def process_delete_records(delete_records):
    """Process the requests for S3 bucket deletions"""
    for r in delete_records:
        arn = "arn:aws:s3:::{}".format(r['detail']['requestParameters']['bucketName'])

        # Need to check if the event is NEWER than the previous event in case
        # events are out of order. This could *possibly* happen if something
        # was deleted, and then quickly re-created. It could be *possible* for the
        # deletion event to arrive after the creation event. Thus, this will check
        # if the current event timestamp is newer and will only delete if the deletion
        # event is newer.
        try:
            print("Deleting bucket: {}".format(arn))
            model = create_delete_model(r)
            model.save(condition=(CurrentS3Model.eventTime <= r["detail"]["eventTime"]))
            model.delete()

        except PynamoDBConnectionError as pdce:
            log.warn("Unable to delete bucket: {}. Either it doesn't exist, or this deletion event is stale "
                     "(arrived before a NEWER creation/update). The specific exception is: {}".format(arn, pdce))


def process_update_records(update_records):
    """Process the requests for S3 bucket update requests"""
    events = sorted(update_records, key=lambda x: x['account'])

    # Group records by account for more efficient processing
    for account_id, events in groupby(events, lambda x: x['account']):
        events = list(events)

        # Grab the bucket names (de-dupe events):
        buckets = {}
        for e in events:
            # If the creation date is present, then use it:
            bucket_event = buckets.get(e["detail"]["requestParameters"]["bucketName"], {
                "creationDate": e["detail"]["requestParameters"].get("creationDate")
            })
            bucket_event.update(e["detail"]["requestParameters"])

            buckets[e["detail"]["requestParameters"]["bucketName"]] = bucket_event
            buckets[e["detail"]["requestParameters"]["bucketName"]]["eventDetails"] = e

        # Query AWS for current configuration
        for b, item in buckets.items():
            print("Processing Create/Update for: {}".format(b))
            # If the bucket does not exist, then simply drop the request --
            # If this happens, there is likely a Delete event that has occurred and will be processed soon.
            try:
                bucket_details = get_bucket(b,
                                            account_number=account_id,
                                            include_created=(item.get("creationDate") is None),
                                            assume_role=HISTORICAL_ROLE,
                                            region=CURRENT_REGION)
                if bucket_details.get("Error"):
                    log.error("[X] Unable to fetch details about bucket: {}. "
                              "The error details are: {}".format(b, bucket_details["Error"]))
                    continue

            except ClientError as ce:
                if ce.response["Error"]["Code"] == "NoSuchBucket":
                    log.warning("[?] Received update request for bucket: {} that does not "
                                "currently exist. Skipping.".format(b))
                    continue

                # Catch Access Denied exceptions as well:
                if ce.response["Error"]["Code"] == "AccessDenied":
                    log.error("[X] Unable to fetch details for S3 Bucket: {} in {}. Access is Denied. Skipping...".format(
                        b, account_id
                    ))
                    continue
                raise Exception(ce)

            # Pull out the fields we want:
            data = {
                "arn": "arn:aws:s3:::{}".format(b),
                "principalId": cloudwatch.get_principal(item["eventDetails"]),
                "userIdentity": cloudwatch.get_user_identity(item["eventDetails"]),
                "accountId": account_id,
                "eventTime": item["eventDetails"]["detail"]["eventTime"],
                "BucketName": b,
                "Region": bucket_details["Region"],
                # Duplicated in top level and configuration for secondary index
                "Tags": bucket_details["Tags"] or {},
                "eventSource": item["eventDetails"]["detail"]["eventSource"]
            }

            # Remove the fields we don't care about:
            del bucket_details["Arn"]
            del bucket_details["GrantReferences"]
            del bucket_details["Region"]

            if not bucket_details.get("CreationDate"):
                bucket_details["CreationDate"] = item["creationDate"]

            data["configuration"] = bucket_details

            current_revision = CurrentS3Model(**data)
            current_revision.save()


@RavenLambdaWrapper()
def handler(event, context):
    """
    Historical S3 event collector.

    This collector is responsible for processing CloudWatch events and polling events.
    """
    records = deserialize_records(event['Records'])

    # Split records into two groups, update and delete.
    # We don't want to query for deleted records.
    update_records, delete_records = group_records_by_type(records)

    log.debug("Processing update records...")
    process_update_records(update_records)
    log.debug("Completed processing of update records.")

    log.debug("Processing delete records...")
    process_delete_records(delete_records)
    log.debug("Completed processing of delete records.")

    log.debug('Successfully updated current Historical table')



"""
.. module: historical.constants
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Mike Grima <mgrima@netflix.com>
.. author:: Kevin Glisson <kglisson@netflix.com>
"""
import os
CURRENT_REGION = os.environ.get('AWS_DEFAULT_REGION', 'us-east-1')
HISTORICAL_ROLE = os.environ.get('HISTORICAL_ROLE', 'Historical')
POLL_REGIONS = os.environ.get('POLL_REGIONS', 'us-east-1').split(",")
DIFF_REGIONS = os.environ.get('DIFF_REGIONS', "us-east-1").split(",")



from jinja2.ext import Extension


def titlecase(f):
    """Transforms a string to titlecase."""
    return "".join([x.title() for x in f.split('_')])


class HistoricalExtension(Extension):
    def __init__(self, environment):
        super(HistoricalExtension, self).__init__(environment)
        environment.filters['titlecase'] = titlecase



"""
.. module: historical.security_group.differ
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
"""
import logging

from raven_python_lambda import RavenLambdaWrapper

from historical.common.dynamodb import process_dynamodb_differ_record
from historical.security_group.models import DurableSecurityGroupModel

logging.basicConfig()
log = logging.getLogger('historical')
log.setLevel(logging.WARNING)


@RavenLambdaWrapper()
def handler(event, context):
    """
    Historical security group event differ.

    Listens to the Historical current table and determines if there are differences that need to be persisted in the
    historical record.
    """
    for record in event['Records']:
        process_dynamodb_differ_record(record, DurableSecurityGroupModel)



"""
.. module: historical.security_group.models
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
"""
from marshmallow import Schema, fields, post_dump

from pynamodb.models import Model
from pynamodb.indexes import GlobalSecondaryIndex, AllProjection
from pynamodb.attributes import UnicodeAttribute, NumberAttribute, ListAttribute

from historical.constants import CURRENT_REGION
from historical.models import (
    HistoricalPollingEventDetail,
    HistoricalPollingBaseModel,
    DurableHistoricalModel,
    CurrentHistoricalModel,
    AWSHistoricalMixin
)


class SecurityGroupModel(object):
    GroupId = UnicodeAttribute()
    GroupName = UnicodeAttribute()
    VpcId = UnicodeAttribute(null=True)
    OwnerId = UnicodeAttribute()
    Description = UnicodeAttribute()
    Tags = ListAttribute()
    Region = UnicodeAttribute()


class DurableSecurityGroupModel(Model, DurableHistoricalModel, AWSHistoricalMixin, SecurityGroupModel):
    class Meta:
        table_name = 'HistoricalSecurityGroupDurableTable'
        region = CURRENT_REGION


class CurrentSecurityGroupModel(Model, CurrentHistoricalModel, AWSHistoricalMixin, SecurityGroupModel):
    class Meta:
        table_name = 'HistoricalSecurityGroupCurrentTable'
        region = CURRENT_REGION


class ViewIndex(GlobalSecondaryIndex):
    class Meta:
        projection = AllProjection()
        region = CURRENT_REGION

    view = NumberAttribute(default=0, hash_key=True)


class SecurityGroupPollingRequestParamsModel(Schema):
    group_id = fields.Str(dump_to='groupId', load_from='groupId', required=True)
    owner_id = fields.Str(dump_to='ownerId', load_from='ownerId', required=True)


class SecurityGroupPollingEventDetail(HistoricalPollingEventDetail):
    @post_dump
    def add_required_security_group_polling_data(self, data):
        data['eventSource'] = 'historical.ec2.poller'
        data['eventName'] = 'HistoricalPoller'
        return data


class SecurityGroupPollingEventModel(HistoricalPollingBaseModel):
    detail = fields.Nested(SecurityGroupPollingEventDetail, required=True)

    @post_dump()
    def dump_security_group_polling_event_data(self, data):
        data['version'] = '1'
        return data

    def serialize(self, account, group):
        return self.dumps({
            'account': account,
            'detail': {
                'request_parameters': {
                    'groupId': group['GroupId']
                }
            }
        }).data


security_group_polling_schema = SecurityGroupPollingEventModel(strict=True)



import sys

import json
import uuid
import base64

import boto3

from historical.constants import CURRENT_REGION


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]


def determine_chunk_size(events):
    """Determines the number of event """
    total_size = 0

    for e in events:
        total_size += sys.getsizeof(e['Data'])
    return total_size


def produce_events(events, stream):
    """
    Efficiently batches and sends events to kinesis stream.

    Initially tries the max batch size of 500 records. If after encoding this
    is over the 5mb limit it attempts to back off the number of records until
    it falls below the 5mb limit.
    """
    client = boto3.client('kinesis', region_name=CURRENT_REGION)

    events = [
        {
            'Data': event.encode('utf-8'),
            'PartitionKey': uuid.uuid4().hex
        } for event in events
    ]

    chunk_size = 500

    while True:
        if determine_chunk_size(events) >= 5242880:
            chunk_size = chunk_size / 2
        else:
            break

    for chunk in chunks(events, chunk_size):
        client.put_records(
            Records=chunk,
            StreamName=stream
        )


def deserialize_records(records):
    """
    Kinesis records come in a base64 encoded format. This function
    parses these records and returns native python data structures.
    """
    native_records = []
    for r in records:
        native_records.append(
            json.loads(
                base64.b64decode(r['kinesis']['data'])
            )
        )
    return native_records



"""
.. module: historical.vpc.collector
    :platform: Unix
    :copyright: (c) 2017 by Netflix Inc., see AUTHORS for more
    :license: Apache, see LICENSE for more details.
.. author:: Kevin Glisson <kglisson@netflix.com>
"""
import os
import logging

from botocore.exceptions import ClientError
from pynamodb.exceptions import DeleteError

from raven_python_lambda import RavenLambdaWrapper

from cloudaux.aws.ec2 import describe_vpcs

from historical.constants import CURRENT_REGION, HISTORICAL_ROLE
from historical.common import cloudwatch
from historical.common.kinesis import deserialize_records
from historical.vpc.models import CurrentVPCModel

logging.basicConfig()
log = logging.getLogger('historical')
level = logging.getLevelName(os.environ.get('HISTORICAL_LOGGING_LEVEL', 'WARNING'))
log.setLevel(level)


UPDATE_EVENTS = [
    'CreateVpc',
    'ModifyVpcAttribute',
    'HistoricalPoller'
]

DELETE_EVENTS = [
    'DeleteVpc'
]


def get_arn(vpc_id, account_id):
    """Creates a vpc ARN."""
    return 'arn:aws:ec2:{region}:{account_id}:vpc/{vpc_id}'.format(
        vpc_id=vpc_id,
        region=CURRENT_REGION,
        account_id=account_id
    )


def group_records_by_type(records):
    """Break records into two lists; create/update events and delete events."""
    update_records, delete_records = [], []
    for r in records:
        if r['detail']['eventName'] in UPDATE_EVENTS:
            update_records.append(r)
        else:
            delete_records.append(r)
    return update_records, delete_records


def describe_vpc(record):
    """Attempts to describe vpc ids."""
    account_id = record['account']
    vpc_name = cloudwatch.filter_request_parameters('vpcName', record)
    vpc_id = cloudwatch.filter_request_parameters('vpcId', record)

    try:
        if vpc_id and vpc_name:
            return describe_vpcs(
                account_number=account_id,
                assume_role=HISTORICAL_ROLE,
                region=CURRENT_REGION,
                Filters=[
                    {
                        'Name': 'vpc-id',
                        'Values': [vpc_id]
                    }
                ]
            )
        elif vpc_id:
            return describe_vpcs(
                account_number=account_id,
                assume_role=HISTORICAL_ROLE,
                region=CURRENT_REGION,
                VpcIds=[vpc_id]
            )
        else:
            raise Exception('Describe requires VpcId.')
    except ClientError as e:
        if e.response['Error']['Code'] == 'InvalidVpc.NotFound':
            return []
        raise e


def create_delete_model(record):
    """Create a vpc model from a record."""
    data = cloudwatch.get_historical_base_info(record)

    vpc_id = cloudwatch.filter_request_parameters('vpcId', record)

    arn = get_arn(vpc_id, record['account'])

    log.debug('Deleting Dynamodb Records. Hash Key: {arn}'.format(arn=arn))

    # tombstone these records so that the deletion event time can be accurately tracked.
    data.update({
        'configuration': {}
    })

    items = list(CurrentVPCModel.query(arn, limit=1))

    if items:
        model_dict = items[0].__dict__['attribute_values'].copy()
        model_dict.update(data)
        model = CurrentVPCModel(**model_dict)
        model.save()
        return model


def capture_delete_records(records):
    """Writes all of our delete events to DynamoDB."""
    for r in records:
        model = create_delete_model(r)
        if model:
            try:
                model.delete(condition=(CurrentVPCModel.eventTime <= r['detail']['eventTime']))
            except DeleteError as _:
                log.warning('Unable to delete VPC. VPC does not exist. Record: {record}'.format(
                    record=r
                ))
        else:
            log.warning('Unable to delete VPC. VPC does not exist. Record: {record}'.format(
                record=r
            ))


def get_vpc_name(vpc):
    """Fetches VPC Name (as tag) from VPC."""
    for t in vpc.get('Tags', []):
        if t['Key'].lower() == 'name':
            return t['Value']


def capture_update_records(records):
    """Writes all updated configuration info to DynamoDB"""
    for record in records:
        data = cloudwatch.get_historical_base_info(record)
        vpc = describe_vpc(record)

        if len(vpc) > 1:
            raise Exception('Multiple vpcs found. Record: {record}'.format(record=record))

        if not vpc:
            log.warning('No vpc information found. Record: {record}'.format(record=record))
            continue

        vpc = vpc[0]

        # determine event data for vpc
        log.debug('Processing vpc. Vpc: {}'.format(vpc))
        data.update({
            'VpcId': vpc.get('VpcId'),
            'Tags': vpc.get('Tags', []),
            'arn': get_arn(vpc['VpcId'], data['accountId']),
            'configuration': vpc,
            'State': vpc.get('State'),
            'IsDefault': vpc.get('IsDefault'),
            'CidrBlock': vpc.get('CidrBlock'),
            'Name': get_vpc_name(vpc),
            'Region': cloudwatch.get_region(record)
        })

        log.debug('Writing Dynamodb Record. Records: {record}'.format(record=data))

        current_revision = CurrentVPCModel(**data)
        current_revision.save()


@RavenLambdaWrapper()
def handler(event, context):
    """
    Historical vpc event collector.
    This collector is responsible for processing Cloudwatch events and polling events.
    """
    records = deserialize_records(event['Records'])

    # Split records into two groups, update and delete.
    # We don't want to query for deleted records.
    update_records, delete_records = group_records_by_type(records)
    capture_delete_records(delete_records)

    # filter out error events
    update_records = [e for e in update_records if not e['detail'].get('errorCode')]

    # group records by account for more efficient processing
    log.debug('Update Records: {records}'.format(records=records))

    capture_update_records(update_records)
